import json

from bson.objectid import ObjectId


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ObjectId):
            return str(o)
        return json.JSONEncoder.default(self, o)



from unittest.mock import Mock
from aiounittest import AsyncTestCase
from py_easy_rest.server import App
from py_easy_rest.repos.memory import MemoryRepo
from py_easy_rest.caches.memory import MemoryCache


api_config_mock = {
    "name": "Project Name",
    "schemas": [{
        "name": "Mock",
        "slug": "mock",
        "properties": {
            "name": {"type": "string"},
            "age": {"type": "integer"},
        },
        "required": ["name"],
    }, {
        "name": "Second api",
        "slug": "second",
        "properties": {"name": {"type": "string"}},
        "enabled_handlers": ["get"]
    }]
}


class BaseSanicTestCase(AsyncTestCase):

    def setUp(self):
        self._mock_repo = Mock(MemoryRepo)
        self._cache = MemoryCache()
        self._py_easy_rest = App(
            api_config_mock,
            repo=self._mock_repo,
            cache=self._cache,
        )

    async def request_api(self, path, method="GET", json=None):
        client = self._py_easy_rest.app.asgi_client

        request, response = await client.request(method, path, json=json)

        await client.aclose()

        return request, response



# Copied from https://stackoverflow.com/questions/20656135/python-deep-merge-dictionary-data
def merge(source, destination):
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            merge(value, node)
        else:
            destination[key] = value

    return destination



import pytest

from aiounittest import AsyncTestCase
from datetime import datetime, timedelta

from py_easy_rest.caches.memory import MemoryCache


class TestMemoryCache(AsyncTestCase):

    @pytest.mark.asyncio
    async def test_should_get_None_when_there_is_no_result_with_key(self):
        cache = MemoryCache()

        result = await cache.get("key")

        assert result is None

    @pytest.mark.asyncio
    async def test_should_get_the_correct_value_when_there_is_result_with_key(self):
        cache = MemoryCache(initial_data={
            "key-1": "value-1",
            "key-2": "value-2",
        })

        result = await cache.get("key-1")

        assert result == "value-1"

    @pytest.mark.asyncio
    async def test_should_get_the_correct_value_when_there_is_result_with_key_and_it_is_not_expired(self):
        cache = MemoryCache(
            initial_data={
                "key-1": "value-1",
                "key-2": "value-2",
            },
            initial_expire_data={
                "key-1": datetime.now() + timedelta(seconds=6)
            }
        )

        result = await cache.get("key-1")

        assert result == "value-1"

    @pytest.mark.asyncio
    async def test_should_get_None_when_there_is_result_with_key_and_it_is_expired(self):
        cache = MemoryCache(
            initial_data={
                "key-1": "value-1",
                "key-2": "value-2",
            },
            initial_expire_data={
                "key-1": datetime.now() + timedelta(seconds=-6)
            }
        )

        result = await cache.get("key-1")

        assert result is None

    @pytest.mark.asyncio
    async def test_should_delete_data_correctly(self):
        cache = MemoryCache(
            initial_data={
                "key-1": "value-1",
                "key-2": "value-2",
            },
        )

        await cache.delete("key-1")

        result = await cache.get("key-1")

        assert result is None

    @pytest.mark.asyncio
    async def test_should_delete_do_nothing_when_there_is_no_result_with_key(self):
        cache = MemoryCache(
            initial_data={
                "key-1": "value-1",
                "key-2": "value-2",
            },
        )

        await cache.delete("key-2")

        result = await cache.get("key-2")

        assert result is None

    @pytest.mark.asyncio
    async def test_should_set_correctly(self):
        cache = MemoryCache()

        await cache.set("key", "value")

        result = await cache.get("key")

        assert result == "value"

    @pytest.mark.asyncio
    async def test_should_set_with_ttl_correctly(self):
        cache = MemoryCache()

        await cache.set("key", "value", 10)

        result = await cache.get("key")

        assert result == "value"



import aioredis

from py_easy_rest.caches import Cache


class RedisCache(Cache):

    def __init__(self, connection_string):
        self._connection_string = connection_string

    async def get(self, key):
        redis = await aioredis.create_redis_pool(self._connection_string)

        value = await redis.get(key, encoding='utf-8')

        redis.close()
        await redis.wait_closed()

        if value:
            return value

        return None

    async def set(self, key, value, ttl=0):
        redis = await aioredis.create_redis_pool(self._connection_string)

        await redis.set(key, value, expire=ttl)

        redis.close()
        await redis.wait_closed()

    async def delete(self, key):
        redis = await aioredis.create_redis_pool(self._connection_string)

        await redis.delete(key)

        redis.close()
        await redis.wait_closed()



from datetime import datetime, timedelta

from py_easy_rest.caches import Cache


class MemoryCache(Cache):

    def __init__(self, initial_data={}, initial_expire_data={}):
        self._data = initial_data
        self._when_data_expire = initial_expire_data

    async def get(self, key):
        value = self._data.get(key)

        if value is not None:
            when_data_expire = self._when_data_expire.get(key)

            if when_data_expire and datetime.now() > when_data_expire:
                await self.delete(key)
                return None

            return value

        return None

    async def set(self, key, value, ttl=None):
        self._data[key] = value

        if ttl is not None:
            when_data_expire = datetime.now() + timedelta(seconds=ttl)
            self._when_data_expire[key] = when_data_expire

    async def delete(self, key):
        self._data.pop(key, None)
        self._when_data_expire.pop(key, None)



from unittest import TestCase

from py_easy_rest.utils.request import get_query_string_arg


class TestRequestUtils(TestCase):

    def test_should_get_arg_return_correct_arg(self):
        query_string = {
            "name": "Jean Pinzon",
        }

        result = get_query_string_arg(query_string, "name")

        assert result == "Jean Pinzon"

    def test_should_get_arg_return_correct_arg_when_it_is_an_array_with_only_one_item(self):
        query_string = {
            "name": ["Jean"],
        }

        result = get_query_string_arg(query_string, "name")

        assert result == "Jean"

    def test_should_get_arg_return_correct_arg_when_it_is_an_array(self):
        query_string = {
            "name": ["Jean", "Alycio"],
        }

        result = get_query_string_arg(query_string, "name")

        assert result == ["Jean", "Alycio"]

    def test_should_get_arg_return_None_when_it_does_not_exists(self):
        query_string = {
            "name": ["Jean", "Alycio"],
        }

        result = get_query_string_arg(query_string, "age")

        assert result is None



def get_query_string_arg(query_string, arg_name):
    arg = query_string.get(arg_name, [])

    if type(arg) is not list:
        return arg

    if len(arg) == 1:
        return arg[0]

    if len(arg) > 1:
        return arg



from bson.objectid import ObjectId

from py_easy_rest.repos import Repo


class MongoRepo(Repo):

    def __init__(self):
        self.connection = None

    def set_db_connection(self, connection):
        self.connection = connection

    async def get(self, slug, id):
        document = await self.connection[slug].find_one({'_id': ObjectId(id)})
        return document

    async def list(self, slug, page, size):
        page = page or 0
        size = size or 30

        cursor = self.connection[slug].find().skip(page * size).limit(size)

        result = await cursor.to_list(length=size)

        return {
            "result": result,
            "page": page,
            "size": size,
        }

    async def create(self, slug, data, id=None):
        if id is not None:
            data['_id'] = ObjectId(id)

        result = await self.connection[slug].insert_one(data)

        return result.inserted_id

    async def replace(self, slug, id, data):
        await self.connection[slug].replace_one({'_id': ObjectId(id)}, data)

    async def delete(self, slug, id):
        await self.connection[slug].delete_one({'_id': ObjectId(id)})



from bson.objectid import ObjectId

from py_easy_rest.repos import Repo


class MemoryRepo(Repo):

    def __init__(self, initial_data=None):
        self._data = {}
        self._data_index = {}

        if initial_data is not None:
            self._data = initial_data

            for key, value in initial_data.items():
                self._data_index[key] = list(self._data[key].keys())

    async def get(self, slug, id):
        self._ensure_slug_exists(slug)
        return self._data[slug].get(id)

    async def list(self, slug, page, size):
        self._ensure_slug_exists(slug)
        page = page or 0
        size = size or 10

        start = page * size
        stop = start + size

        result = self._data_index[slug][start:stop]

        return {
            "result": [self._data[slug][id] for id in result],
            "page": page,
            "size": size,
            "totalCount": len(self._data_index[slug])
        }

    async def create(self, slug, data, id=None):
        self._ensure_slug_exists(slug)
        if id is None:
            id = ObjectId()

        if id not in self._data_index[slug]:
            self._data_index[slug].append(id)

        self._data[slug][id] = data

        return id

    async def replace(self, slug, id, data):
        self._ensure_slug_exists(slug)
        if id in self._data_index[slug]:
            self._data[slug][id] = data

    async def delete(self, slug, id):
        self._ensure_slug_exists(slug)
        self._data[slug].pop(id, None)

        if id in self._data_index[slug]:
            self._data_index[slug].remove(id)

    def _ensure_slug_exists(self, slug):
        if not self._data.get(slug):
            self._data[slug] = {}

        if not self._data_index.get(slug):
            self._data_index[slug] = []



"""
Module with cache providers to be used connected with api.
"""

from py_easy_rest import PYRApplicationError


class CacheError(PYRApplicationError):
    """
    Error to raise in case of database errors.
    """

    pass


class Cache():
    """
    Interface to define contract to cache providers.
    All methods raises an <CacheError> in case of error.
    """

    async def get(self, key):
        """
        Receives <key> and return a string with the result.
        If it not found a data with this <key>, return None.
        """
        raise NotImplementedError

    async def delete(self, key):
        """
        Receives <key> and delete it from cache.
        """
        raise NotImplementedError

    async def set(self, key, value, ttl=None):
        """
        Receives <key> and <value> and set it into cache.
        If it receives a <ttl>, the value is setted with it.
        <ttl> is the time to the cache expires in seconds.
        """
        raise NotImplementedError



"""
Module with repositories to be used connected with api.
"""

from py_easy_rest import PYRApplicationError


class DatabaseError(PYRApplicationError):
    """
    Error to raise in case of database errors.
    """

    pass


class Repo():
    """
    Interface to define contract to repositories.
    All methods raises an <DatabaseError> in case of error.
    """

    async def get(self, slug, id):
        """
        Receives <slug> and <id> and return a dictionary with the result.
        If it not found a data with this <id>, return None.
        """
        raise NotImplementedError

    async def list(self, slug, page=0, size=10):
        """
        Receives <slug>, <page> and <size> and return a object with the result, page and size.
        It's possible to put other properties in this result object too.
        If result is empty, return a empty list.
        """
        raise NotImplementedError

    async def create(self, slug, data, id=None):
        """
        Receives <slug> and <data> with the resource to be saved,
        save it into db, and return the resource id.
        If receive <id>, set this id to the document and return this id.
        """
        raise NotImplementedError

    async def replace(self, slug, id, data):
        """
        Receives <slug> and <data> with the resource to be saved and <id>,
        replace the resource in db.
        """
        raise NotImplementedError

    async def delete(self, slug, id):
        """
        Receives <slug> and <id> and delete it from db.
        """
        raise NotImplementedError



import pytest

from aiounittest import AsyncTestCase

from py_easy_rest.caches.dummy import DummyCache


class TestDummyCache(AsyncTestCase):

    def setUp(self):
        self._cache = DummyCache()

    @pytest.mark.asyncio
    async def test_should_get_return_None(self):
        result = await self._cache.get("mock_key")

        assert result is None

    @pytest.mark.asyncio
    async def test_should_delete_return_None(self):
        result = await self._cache.delete("mock_key")

        assert result is None

    @pytest.mark.asyncio
    async def test_should_set_return_None(self):
        result = await self._cache.set("mock_key", "mock_value")

        assert result is None




import json

from jsonschema import Draft7Validator
from sanic import Sanic, response
from sanic.log import logger
from sanic_openapi import doc, swagger_blueprint

from py_easy_rest import PYRApplicationError
from py_easy_rest.caches.dummy import DummyCache
from py_easy_rest.repos.memory import MemoryRepo
from py_easy_rest.utils.dictionary import merge
from py_easy_rest.utils.json import JSONEncoder
from py_easy_rest.utils.request import get_query_string_arg


json_dumps = JSONEncoder().encode


class App():

    def __init__(
        self,
        api_config,
        repo=MemoryRepo(),
        cache=DummyCache(),
        cache_list_seconds_ttl=10,
        cache_get_seconds_ttl=60 * 30,  # thirty minutes
    ):
        self._repo = repo
        self._api_config = api_config
        self._cache = cache
        self._cache_list_seconds_ttl = cache_list_seconds_ttl
        self._cache_get_seconds_ttl = cache_get_seconds_ttl

        self._schemas = self._api_config["schemas"]

        self.app = Sanic(self._api_config["name"])

        self.app.blueprint(swagger_blueprint)

        self.app.config["API_TITLE"] = self._api_config["name"]

        self.app.error_handler.add(PYRApplicationError, App._handle_app_error)

        for schema in self._schemas:
            self._define_routes(schema)

    @staticmethod
    async def _handle_app_error(request, exception):
        logger.exception(f"Failed to handle request {exception}")
        return response.json({"message": exception.user_message}, status=500, dumps=json_dumps)

    def _validate(self, resource, schema):
        validator = Draft7Validator(schema)

        errors = []

        for error in validator.iter_errors(resource):
            errors.append(error.message)

        if len(errors) > 0:
            return errors

        return None

    def _define_routes(self, schema):
        slug = schema['slug']
        name = schema['name']
        enabled_handlers = schema.get('enabled_handlers', [
            "list",
            "create",
            "get",
            "replace",
            "partial_update",
            "delete",
        ])

        @self.app.get(f"/{slug}/schema")
        @doc.summary("Get JSON Schema")
        @doc.description("Route to get the api JSON Schema.")
        @doc.response(200, None, description="Success to get JSON Schema.")
        async def _get_schema(request):
            return response.json(schema, dumps=json_dumps)

        if "list" in enabled_handlers:
            @self.app.get(f"/{slug}")
            @doc.summary(f"List {name}")
            @doc.description(
                f"Route to list all {name}. "
                "You can use the parameters page and size. Default values: page=0, size=30."
            )
            @doc.response(200, None, description=f"Success to list {name}.")
            @doc.response(500, None, description="Internal server error.")
            @doc.consumes(doc.Integer(name="page"), location="query")
            @doc.consumes(doc.Integer(name="size"), location="query")
            @doc.produces(doc.List(doc.Dictionary()))
            async def _list(request):
                page = get_query_string_arg(request.args, "page")
                size = get_query_string_arg(request.args, "size")

                if page is not None:
                    page = int(page)

                if size is not None:
                    size = int(size)

                cache_key = f"{slug}.list.page-{page}.size-{size}"

                cached = await self._cache.get(cache_key)

                if cached is not None:
                    logger.info(f"Found cache result with key {cache_key}")
                    result = json.loads(cached)
                    return response.json(result, dumps=json_dumps)

                logger.info(f"Not found cache result with key {cache_key}")

                result = await self._repo.list(slug, page, size)

                await self._cache.set(cache_key, json_dumps(result), ttl=self._cache_list_seconds_ttl)

                return response.json(result, dumps=json_dumps)

        if "create" in enabled_handlers:
            @self.app.post(f"/{slug}")
            @self.app.post(f"/{slug}/<id>")
            @doc.summary(f"Create a new {name}")
            @doc.description(
                f"Route to create a new document of {name}. "
                "Take a look at the schema route to know what properties you must send."
            )
            @doc.response(201, None, description=f"Success to create a new {name} document.")
            @doc.response(400, None, description="Validation error.")
            @doc.response(500, None, description="Internal server error.")
            @doc.consumes(doc.JsonBody({}), location="body")
            async def _post(request, id=None):
                errors = self._validate(request.json, schema)

                if errors:
                    return response.json({"errors": errors}, status=400, dumps=json_dumps)

                resource_id = await self._repo.create(slug, request.json, id)

                if id:
                    cache_key = f"{slug}.get.id-{id}"
                    await self._cache.delete(cache_key)

                return response.json({"id": resource_id}, status=201, dumps=json_dumps)

        if "get" in enabled_handlers:
            @self.app.get(f"/{slug}/<id>")
            @doc.summary("Get a document by id")
            @doc.description(f"Route to get a {name} document by id.")
            @doc.response(200, None, description="Success to get the document.")
            @doc.response(404, None, description="Document not found.")
            @doc.response(500, None, description="Internal server error.")
            @doc.produces(doc.Dictionary())
            async def _get(request, id):
                cache_key = f"{slug}.get.id-{id}"

                cached = await self._cache.get(cache_key)

                if cached is not None:
                    logger.info(f"Found cache result with key {cache_key}")
                    result = json.loads(cached)
                    return response.json(result, dumps=json_dumps)

                logger.info(f"Not found cache result with key {cache_key}")

                result = await self._repo.get(slug, id)

                if result:
                    await self._cache.set(cache_key, json_dumps(result), ttl=self._cache_get_seconds_ttl)
                    return response.json(result, dumps=json_dumps)

                return response.json({}, status=404, dumps=json_dumps)

        if "replace" in enabled_handlers:
            @self.app.put(f"/{slug}/<id>")
            @doc.summary("Replace a document by id")
            @doc.description(
                f"Route to replace a document of {name}. "
                "Take a look at the schema route to know what properties you must send."
            )
            @doc.response(200, None, description="Success to replace the document.")
            @doc.response(400, None, description="Validation error.")
            @doc.response(500, None, description="Internal server error.")
            @doc.consumes(doc.JsonBody({}), location="body")
            async def _put(request, id):
                existent_doc = await self._repo.get(slug, id)

                if not existent_doc:
                    return response.json({}, status=404, dumps=json_dumps)

                errors = self._validate(request.json, schema)

                if errors:
                    return response.json({"errors": errors}, status=400, dumps=json_dumps)

                await self._repo.replace(slug, id, request.json)

                cache_key = f"{slug}.get.id-{id}"
                await self._cache.delete(cache_key)

                return response.json({}, dumps=json_dumps)

        if "partial_update" in enabled_handlers:
            @self.app.patch(f"/{slug}/<id>")
            @doc.summary("Partial update a document by id")
            @doc.description(
                f"Route to partial update a document of {name}. "
                "Take a look at the schema route to know what properties you must send."
            )
            @doc.response(200, None, description="Success to partial update the document.")
            @doc.response(400, None, description="Validation error.")
            @doc.response(500, None, description="Internal server error.")
            @doc.consumes(doc.JsonBody({}), location="body")
            async def _patch(request, id):
                existent_doc = await self._repo.get(slug, id)

                if not existent_doc:
                    return response.json({}, status=404, dumps=json_dumps)

                doc = merge(request.json, existent_doc)
                doc.pop("_id", None)

                errors = self._validate(doc, schema)

                if errors:
                    return response.json({"errors": errors}, status=400, dumps=json_dumps)

                await self._repo.replace(slug, id, doc)

                cache_key = f"{slug}.get.id-{id}"
                await self._cache.delete(cache_key)

                return response.json({}, dumps=json_dumps)

        if "delete" in enabled_handlers:
            @self.app.delete(f"/{slug}/<id>")
            @doc.summary("Delere a document by id")
            @doc.description(f"Route to delete a document of {name}.")
            @doc.response(200, None, description="Success to delete the document.")
            @doc.response(500, None, description="Internal server error.")
            async def _delete(request, id):
                existent_doc = await self._repo.get(slug, id)

                if not existent_doc:
                    return response.json({}, status=404, dumps=json_dumps)

                await self._repo.delete(slug, id)

                cache_key = f"{slug}.get.id-{id}"
                await self._cache.delete(cache_key)

                return response.json({}, dumps=json_dumps)



import pytest
import json

from py_easy_rest.repos import DatabaseError

from tests.py_easy_rest import BaseSanicTestCase


class TestServer(BaseSanicTestCase):

    @pytest.mark.asyncio
    async def test_should_get_with_a_not_existent_name_returns_404(self):
        request, response = await self.request_api("/not-existent-name")

        assert response.status == 404

    @pytest.mark.asyncio
    async def test_should_get_schema_returns_200(self):
        expected_schema = {
            "name": "Mock",
            "slug": "mock",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "required": ["name"],
        }

        request, response = await self.request_api("/mock/schema")

        assert response.status == 200
        assert response.json() == expected_schema

    @pytest.mark.asyncio
    async def test_should_works_with_multiple_schemas_correctly(self):
        expected_schema = {
            "name": "Second api",
            "slug": "second",
            "properties": {"name": {"type": "string"}},
            "enabled_handlers": ["get"]
        }

        request, response = await self.request_api("/second/schema")

        assert response.status == 200
        assert response.json() == expected_schema

    @pytest.mark.asyncio
    async def test_should_get_without_id_returns_200_and_a_list_of_resources(self):
        expected_list_of_resources = [
            {"name": "Jean Pinzon"},
            {"name": "Alycio Neto"},
        ]

        self._mock_repo.list.return_value = expected_list_of_resources

        await self._cache.delete("mock.list.page-None.size-None")

        request, response = await self.request_api("/mock")

        assert response.status == 200
        assert response.json() == expected_list_of_resources

        self._mock_repo.list.assert_called_once()

    @pytest.mark.asyncio
    async def test_should_get_without_id_returns_200_and_a_cached_list(self):
        cached_list = [
            {"name": "Jean Pinzon"},
            {"name": "Alycio Neto"},
        ]

        await self._cache.set("mock.list.page-None.size-None", json.dumps(cached_list))

        request, response = await self.request_api("/mock")

        assert response.status == 200
        assert response.json() == cached_list

    @pytest.mark.asyncio
    async def test_should_get_without_id_pass_pagination_params_to_repo_when_receives_it_as_query_string(self):
        expected_page = 10
        expected_size = 20
        expected_list_of_resources = []

        self._mock_repo.list.return_value = []

        request, response = await self.request_api(f"/mock?page={expected_page}&size={expected_size}")

        assert response.status == 200
        assert response.json() == expected_list_of_resources

        self._mock_repo.list.assert_called_once_with("mock", expected_page, expected_size)

    @pytest.mark.asyncio
    async def test_should_get_with_id_returns_200_and_a_resource(self):
        expected_resource = {"name": "Jean Pinzon"}

        self._mock_repo.get.return_value = expected_resource

        request, response = await self.request_api("/mock/1")

        assert response.status == 200
        assert response.json() == expected_resource

    @pytest.mark.asyncio
    async def test_should_get_with_id_returns_200_and_cached_resource(self):
        cached_resource = {"name": "Jean Pinzon"}

        await self._cache.set("mock.get.id-6", json.dumps(cached_resource))

        request, response = await self.request_api("/mock/6")

        assert response.status == 200
        assert response.json() == cached_resource

    @pytest.mark.asyncio
    async def test_should_get_with_id_returns_404_if_resource_not_found(self):
        expected_resource = None

        self._mock_repo.get.return_value = expected_resource

        request, response = await self.request_api("/mock/2")

        assert response.status == 404

    @pytest.mark.asyncio
    async def test_should_post_returns_201_and_resource_id(self):
        resource = {"name": "karl"}
        expected_id = 'mock-id'

        self._mock_repo.create.return_value = expected_id

        request, response = await self.request_api(
            path="/mock",
            method="POST",
            json=resource
        )

        assert response.status == 201
        assert response.json() == {"id": expected_id}

        self._mock_repo.create.assert_called_once_with("mock", resource, None)

    @pytest.mark.asyncio
    async def test_should_post_returns_400_and_errors_when_data_is_not_valid(self):
        resource = {"name": "karl", "age": "twenty eight"}
        expected_id = 'mock-id'

        self._mock_repo.create.return_value = expected_id

        request, response = await self.request_api(
            path="/mock",
            method="POST",
            json=resource
        )

        assert response.status == 400
        assert response.json() == {"errors": ["'twenty eight' is not of type 'integer'"]}

        self._mock_repo.create.assert_not_called()

    @pytest.mark.asyncio
    async def test_should_post_with_id_returns_201_and_resource_id(self):
        resource = {"name": "karl"}
        resource_id = 'mock-id'

        self._mock_repo.create.return_value = resource_id

        request, response = await self.request_api(
            path=f"/mock/{resource_id}",
            method="POST",
            json=resource
        )

        assert response.status == 201
        assert response.json() == {"id": resource_id}

        self._mock_repo.create.assert_called_once_with("mock", resource, resource_id)

    @pytest.mark.asyncio
    async def test_should_put_returns_200(self):
        resource = {"name": "karl"}
        resource_id = 'mock-id'

        request, response = await self.request_api(
            path=f"/mock/{resource_id}",
            method="PUT",
            json=resource
        )

        assert response.status == 200
        assert response.json() == {}

        self._mock_repo.replace.assert_called_once_with("mock", resource_id, resource)

    @pytest.mark.asyncio
    async def test_should_put_returns_404(self):
        resource = {"name": "karl"}
        resource_id = 'not-found-id-put'

        self._mock_repo.get.return_value = None

        request, response = await self.request_api(
            path=f"/mock/{resource_id}",
            method="PUT",
            json=resource
        )

        assert response.status == 404
        assert response.json() == {}

        self._mock_repo.replace.assert_not_called()

    @pytest.mark.asyncio
    async def test_should_put_returns_400_and_errors_when_data_is_not_valid(self):
        resource = {"name": "karl", "age": "twenty eight"}
        resource_id = 'mock-id'

        request, response = await self.request_api(
            path=f"/mock/{resource_id}",
            method="PUT",
            json=resource
        )

        assert response.status == 400
        assert response.json() == {"errors": ["'twenty eight' is not of type 'integer'"]}

        self._mock_repo.create.assert_not_called()

    @pytest.mark.asyncio
    async def test_should_patch_returns_200(self):
        resource = {"name": "karl"}
        resource_id = 'mock-id'

        self._mock_repo.get.return_value = {"name": "jean"}

        request, response = await self.request_api(
            path=f"/mock/{resource_id}",
            method="PATCH",
            json=resource
        )

        assert response.status == 200
        assert response.json() == {}

        self._mock_repo.replace.assert_called_once_with("mock", resource_id, resource)

    @pytest.mark.asyncio
    async def test_should_patch_returns_404(self):
        resource = {"name": "karl"}
        resource_id = 'not-found-id-patch'

        self._mock_repo.get.return_value = None

        request, response = await self.request_api(
            path=f"/mock/{resource_id}",
            method="PATCH",
            json=resource
        )

        assert response.status == 404
        assert response.json() == {}

        self._mock_repo.replace.assert_not_called()

    @pytest.mark.asyncio
    async def test_should_patch_returns_400_and_errors_when_data_is_not_valid(self):
        resource = {"name": "karl", "age": "twenty eight"}
        resource_id = 'mock-id'

        self._mock_repo.get.return_value = {"name": "jean"}

        request, response = await self.request_api(
            path=f"/mock/{resource_id}",
            method="PATCH",
            json=resource
        )

        assert response.status == 400
        assert response.json() == {"errors": ["'twenty eight' is not of type 'integer'"]}

        self._mock_repo.create.assert_not_called()

    @pytest.mark.asyncio
    async def test_should_delete_returns_200(self):
        resource_id = 'mock-id'

        request, response = await self.request_api(
            path=f"/mock/{resource_id}",
            method="DELETE",
        )

        assert response.status == 200
        assert response.json() == {}

        self._mock_repo.delete.assert_called_once_with("mock", resource_id)

    @pytest.mark.asyncio
    async def test_should_delete_returns_404(self):
        resource = {"name": "karl"}
        resource_id = 'not-found-id-delete'

        self._mock_repo.get.return_value = None

        request, response = await self.request_api(
            path=f"/mock/{resource_id}",
            method="DELETE",
            json=resource
        )

        assert response.status == 404
        assert response.json() == {}

        self._mock_repo.replace.assert_not_called()

    @pytest.mark.asyncio
    async def test_should_request_returns_500_when_it_result_in_a_not_expected_error(self):
        self._mock_repo.list.side_effect = Exception()

        request, response = await self.request_api("/mock?page=321")

        assert response.status == 500

        self._mock_repo.list.assert_called_once()

    @pytest.mark.asyncio
    async def test_should_request_with_id_returns_500_when_it_result_in_a_not_expected_error(self):
        self._mock_repo.get.side_effect = Exception()

        request, response = await self.request_api("/mock/234")

        assert response.status == 500

    @pytest.mark.asyncio
    async def test_should_request_returns_500_and_message_when_it_result_in_a_db_error(self):
        expected_error_message = "User friendly error message"

        self._mock_repo.list.side_effect = DatabaseError(
            error_message="System error!",
            user_message=expected_error_message
        )

        request, response = await self.request_api("/mock?page=123")

        assert response.status == 500
        assert response.json() == {"message": expected_error_message}

        self._mock_repo.list.assert_called_once()

    @pytest.mark.asyncio
    async def test_should_request_with_id_returns_500_and_message_when_it_result_in_a_db_error(self):
        resource_id = 'mock-id'
        expected_error_message = "User friendly error message"

        self._mock_repo.get.side_effect = DatabaseError(
            error_message="System error!",
            user_message=expected_error_message
        )

        request, response = await self.request_api(f"/mock/{resource_id}")

        assert response.status == 500
        assert response.json() == {"message": expected_error_message}

        self._mock_repo.get.assert_called_once_with("mock", resource_id)

    @pytest.mark.asyncio
    async def test_should_disabled_handlers_return_405(self):
        request, response = await self.request_api("/second/1", method="DELETE")

        assert response.status == 405



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="py-easy-rest",
    version="0.3.0",
    author="Jean Pinzon",
    author_email="jean.pinzon1@gmail.com",
    description="The simplest way to have a rest api",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(
        exclude=(
            'tests',
        ),
    ),
    install_requires=[
        "aioredis==1.3.1",
        "dnspython==2.1.0",
        "jsonschema==3.2.0",
        "motor==2.3.0",
        "sanic-openapi==0.6.2",
        "sanic==20.9.1",
    ],
    extras_require={
        'tests': [
            "pytest==6.1.2",
            "pytest-asyncio==0.14.0",
            "pytest-cov==2.10.1",
            "flake8==3.8.4",
            "aiounittest==1.4.0",
            "twine==3.3.0",
            "build==0.3.1.post1",
        ],
    },
    python_requires='>=3.8',
)



from py_easy_rest.caches import Cache


class DummyCache(Cache):

    async def get(self, key):
        return None

    async def delete(self, key):
        return None

    async def set(self, key, value, ttl=None):
        return None



class PYRApplicationError(Exception):
    """
    Error to raise in case of now you rest application errors.
    """

    def __init__(self, error_message, user_message=None, extra_params={}):
        self.error_message = error_message
        self.user_message = user_message
        self.extra_params = extra_params



from unittest import TestCase

from py_easy_rest.utils.dictionary import merge


class TestDictionaryUtils(TestCase):

    def test_should_merge_correctly(self):
        dict_a = {'first': {'all_rows': {'pass': 'dog', 'number': '1'}}}
        dict_b = {'first': {'all_rows': {'fail': 'cat', 'number': '5'}}}

        expected_dict = {'first': {'all_rows': {'pass': 'dog', 'fail': 'cat', 'number': '5'}}}

        result = merge(dict_b, dict_a)

        assert result == expected_dict



from unittest import TestCase

from bson.objectid import ObjectId

from py_easy_rest.utils.json import JSONEncoder


class TestJsonUtils(TestCase):

    json_dumps = JSONEncoder().encode

    def test_should_parse_object_id_to_string(self):
        result = self.json_dumps({"id": ObjectId("551137c2f9e1fac808a5f572")})

        assert result == '{"id": "551137c2f9e1fac808a5f572"}'
